package ru.dbotthepony.mc.otm.block.entity.matter

import net.minecraft.core.BlockPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.block.matter.MatterBottlerBlock
import ru.dbotthepony.mc.otm.block.entity.MatteryPoweredBlockEntity
import ru.dbotthepony.mc.otm.block.entity.WorkerState
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.item.ProxiedItemHandler
import ru.dbotthepony.mc.otm.capability.UpgradeType
import ru.dbotthepony.mc.otm.capability.energy.ProfiledEnergyStorage
import ru.dbotthepony.mc.otm.capability.energy.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.MatterStorageImpl
import ru.dbotthepony.mc.otm.capability.matter.ProfiledMatterStorage
import ru.dbotthepony.mc.otm.config.MachinesConfig
import ru.dbotthepony.mc.otm.container.UpgradeContainer
import ru.dbotthepony.mc.otm.container.slotted.ContainerSlot
import ru.dbotthepony.mc.otm.container.slotted.FilteredContainerSlot
import ru.dbotthepony.mc.otm.container.slotted.SlottedContainer
import ru.dbotthepony.mc.otm.menu.matter.MatterBottlerMenu
import ru.dbotthepony.mc.otm.registry.game.MBlockEntities
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.util.countingLazy
import ru.dbotthepony.mc.otm.graph.matter.SimpleMatterNode
import java.util.function.BooleanSupplier

class MatterBottlerBlockEntity(blockPos: BlockPos, blockState: BlockState) :
	MatteryPoweredBlockEntity(MBlockEntities.MATTER_BOTTLER, blockPos, blockState) {

	val upgrades = UpgradeContainer(3, UpgradeType.BASIC_MATTER, BooleanSupplier { blockState.getValue(WorkerState.SEMI_WORKER_STATE) !== WorkerState.IDLE }, ::markDirtyFast)
	override val energy = ProfiledEnergyStorage(WorkerEnergyStorage(this::markDirtyFast, upgrades.transform(MachinesConfig.MatterBottler.VALUES)))
	val energyConfig = ConfigurableEnergy(energy)

	private inner class BottlingSlot(container: SlottedContainer, slot: Int) : FilteredContainerSlot(container, slot) {
		override fun canAutomationPlaceItem(itemStack: ItemStack): Boolean {
			return super.canAutomationPlaceItem(itemStack) && isBottling && itemStack.getCapability(MatteryCapability.MATTER_ITEM)?.receiveMatterChecked(Decimal.ONE, true)?.isPositive == true
		}
	}

	private inner class UnBottlingSlot(container: SlottedContainer, slot: Int) : FilteredContainerSlot(container, slot) {
		override fun canAutomationPlaceItem(itemStack: ItemStack): Boolean {
			return super.canAutomationPlaceItem(itemStack) && !isBottling && itemStack.getCapability(MatteryCapability.MATTER_ITEM)?.extractMatterChecked(Decimal.ONE, true)?.isPositive == true
		}
	}

	val unbottling = SlottedContainer.simple(3, ::UnBottlingSlot, ::updateBlockState)
	val bottling = SlottedContainer.simple(3, ::BottlingSlot, ::updateBlockState)

	var isBottling: Boolean = true
		set(value) {
			field = value
			initialCapacity = null
			workProgress = 0f
			this.markDirtyFast()

			if (value) {
				inputHandler.parent = bottling
				outputHandler.parent = unbottling
			} else {
				inputHandler.parent = unbottling
				outputHandler.parent = bottling
			}

			updateBlockState()
		}

	var spitItemsWhenCantWork = false
		set(value) {
			field = value
			this.markDirtyFast()
		}

	private val inputHandler = ProxiedItemHandler(bottling)
	private val outputHandler = ProxiedItemHandler(unbottling)

	val itemConfig = ConfigurableItemHandler(
		input = inputHandler,
		output = outputHandler,
		battery = batteryContainer
	)

	val matter: ProfiledMatterStorage<MatterStorageImpl> = ProfiledMatterStorage(object : MatterStorageImpl(this@MatterBottlerBlockEntity::markDirtyFast, FlowDirection.BI_DIRECTIONAL, upgrades.matterCapacity(MachinesConfig.MatterBottler.VALUES::matterCapacity)) {
		override val matterFlow: FlowDirection get() {
			return if (this@MatterBottlerBlockEntity.isBottling) FlowDirection.INPUT else FlowDirection.OUTPUT
		}
	})

	val matterNode = SimpleMatterNode(matter = matter)

	init {
		exposeGlobally(MatteryCapability.MATTER_BLOCK, matter)
		exposeGlobally(MatteryCapability.MATTER_NODE, matterNode)

		savetablesConfig.bool(::isBottling)
		savetablesConfig.bool(::spitItemsWhenCantWork)
		savetables.stateful(::energy, ENERGY_KEY)
		savetables.stateful(::matter, MATTER_STORAGE_KEY)
		savetables.stateful(::bottling)
		savetables.stateful(::unbottling)
		savetables.stateful(::upgrades)
	}

	var workProgress: Float = 0f
		private set

	private var lastTickWorking = false
	private var initialCapacity: Decimal? = null

	override fun setLevel(level: Level) {
		super.setLevel(level)
		matterNode.discover(this)
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterBottlerMenu(containerID, inventory, this)
	}

	override fun setRemoved() {
		super.setRemoved()
		matterNode.isValid = false
	}

	private fun updateBlockState(container: SlottedContainer) {
		val level = level as? ServerLevel ?: return
		var state = blockState

		for ((i, slot) in container.slotIterator().withIndex()) {
			val desired = !slot.isEmpty && slot.item.getCapability(MatteryCapability.MATTER_ITEM) != null

			if (state.getValue(MatterBottlerBlock.SLOT_PROPERTIES[i]) != desired) {
				state = state.setValue(MatterBottlerBlock.SLOT_PROPERTIES[i], desired)
			}
		}

		if (state !== blockState) {
			level.setBlock(blockPos, state, Block.UPDATE_CLIENTS)
		}
	}

	private fun updateBlockState() {
		markDirtyFast()
		if (isBottling)
			updateBlockState(bottling)
		else
			updateBlockState(unbottling)
	}

	private val workerState by countingLazy(blockStateChangesCounter) {
		blockState.getValue(WorkerState.SEMI_WORKER_STATE)
	}

	private fun blockstateToWorking() {
		if (workerState !== WorkerState.WORKING) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS)
		}
	}

	private fun blockstateToIdle() {
		if (workerState !== WorkerState.IDLE) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
		}
	}

	override fun tick() {
		super.tick()

		if (redstoneControl.isBlockedByRedstone) {
			blockstateToIdle()
			return
		}

		if (isBottling) {
			var any = false
			var idle = false

			for (slot in bottling.slotIterator()) {
				val item = slot.item

				item.getCapability(MatteryCapability.MATTER_ITEM)?.let {
					if (!it.missingMatter.isPositive) {
						unbottling.consumeItem(item, false)
						slot.setChanged()
					} else {
						any = true
						initialCapacity = initialCapacity ?: it.storedMatter
						val rate = MachinesConfig.MatterBottler.RATE * (1.0 + upgrades.speedBonus)

						if (matter.storedMatter < rate) {
							matter.receiveMatter(matterNode.graph.extractMatter(matter.missingMatter
									.coerceAtMost(rate * 200)
									.coerceAtMost(it.missingMatter - matter.storedMatter), false), false)
						}

						if (matter.storedMatter.isPositive) {
							matter.extractMatter(it.receiveMatter(rate.coerceAtMost(matter.storedMatter), false), false)

							if (!it.missingMatter.isPositive) {
								initialCapacity = null
								workProgress = 0f
							} else {
								workProgress = ((it.storedMatter - initialCapacity!!) / it.maxStoredMatter).toFloat()
							}
						} else {
							idle = true

							if (spitItemsWhenCantWork) {
								unbottling.consumeItem(item, false)
								slot.setChanged()
							}
						}
					}
				}

				if (any) {
					break
				}
			}

			if (any && !idle) {
				blockstateToWorking()
			} else {
				matter.extractMatter(matterNode.graph.receiveMatter(matter.storedMatter, false), false)
				blockstateToIdle()
			}
		} else {
			matter.extractMatter(matterNode.graph.receiveMatter(matter.storedMatter, false), false)

			if (!matter.missingMatter.isPositive) {
				if (spitItemsWhenCantWork) {
					for (slot in unbottling.slotIterator()) {
						bottling.consumeItem(slot.item, false)
						slot.setChanged()
					}
				}

				blockstateToIdle()
				return
			}

			var any = false

			for (slot in unbottling.slotIterator()) {
				val item = slot.item

				item.getCapability(MatteryCapability.MATTER_ITEM)?.let {
					if (!it.storedMatter.isPositive) {
						bottling.consumeItem(item, false)
						slot.setChanged()
					} else {
						any = true
						initialCapacity = initialCapacity ?: it.storedMatter
						matter.receiveMatter(it.extractMatter(MachinesConfig.MatterBottler.RATE, false), false)

						if (!it.storedMatter.isPositive) {
							initialCapacity = null
							workProgress = 0f
						} else {
							workProgress = 1f - (it.storedMatter / initialCapacity!!).toFloat()
						}
					}
				}

				if (any) {
					break
				}
			}

			if (any) {
				blockstateToWorking()
			} else {
				blockstateToIdle()
			}
		}
	}
}
